program Test_DNNL;
(*******************************************************************************
 * Sample application demonstrating usage of Intel® oneAPI Deep Neural Network Library (oneDNNL) C API
 *
 * TCNN (Convolutional Neural Network) class provides sample methods for:
 * - creating layers
 * - forward propagation
 * - backward propagation
 *
 * Example:
 *   var
 *     cnn: TCNN;
 *   begin
 *     cnn := TCNN.Create;
 *     // #0: add input layer (gray-scale image): 1x6x6
 *     cnn.AddLayer(cltInput, 1{channel}, W, H, 0,0,0);
 *     // #1: add 1. convolution layer: 8x6x6
 *     cnn.AddLayer(cltConvolution, 8{depth}, 6{width}, 6{height}, 3{kernel}, 1{stride}, 1{padding});
 *     // #2: add ReLU layer
 *     cnn.AddLayer(cltFuncReLU, 8{depth}, 6{width}, 6{height}, 0,0,0);
 *     // #3: add 1. MaxPool layer: 8x3x3
 *     cnn.AddLayer(cltMaxPooling, 8{depth}, 3{width}, 3{height}, 2{kernel}, 2{stride}, 0);
 *     // #4: add 1. fully connected layer
 *     cnn.AddLayer(cltFullyConnected, 10{depth/neurons}, 1{width}, 1{height}, 1{kernel}, 1{stride}, 0{padding});
 *
 *     cnn.Initialize;
 *
 *     ...
 *     cnn.ForwardPropagation;
 *     ...
 *     cnn.BackwardPropagation;
 *     ...
 *     cnn.Free;
 *   end;
 ******************************************************************************)
{$mode objfpc}{$H+}

uses
  {$IFDEF UNIX}
  cthreads,
  {$ENDIF}
  Classes,
  SysUtils,
  dnnl;

{$DEFINE HAVE_TCNN} // does we want TCNN wrapper class?
{ $DEFINE DEBUG_TCNN}

const
  // Input grayscale image W x H x 1
  W=6;
  H=6;
  input_src_sizes: dnnl_dims_t = (1{n},1{c},H{h},W{w},0,0,0,0,0,0,0,0);
  input_: array[0..W*H-1] of single = (
   1, 1, 1, 1, 1, 1,
   1, 2, 1, 1, 1, 1,
   1, 1, 1, 1, 1, 1,
   1, 1, 1, 1, 1, 1,
   1, 1, 1, 1, 1, 1,
   1, 1, 1, 1, 1, 1
  );

type
  TCNNLayerType = (cltInput, cltConvolution, cltFuncReLU, cltMaxPooling, cltAvgPooling, cltFullyConnected, cltFuncTanh, cltFuncLogSig);

  TCNNPrimitive = record
    PrimitiveDesc: dnnl_primitive_desc_t;
    Primitive: dnnl_primitive_t;
    nargs: integer;
    args: array[0..3] of dnnl_exec_arg_t;
  end;

  TCNNPrimitives = array[0..3] of TCNNPrimitive;

  { TCNNLayer }

  TCNNLayer = class
    private
      FLayerType: TCNNLayerType;
      FDepth, FHeight, FWidth: integer; // dimensions of feature maps (C x H x W)
      FKernelSize: integer;  // size of convolution kernel (filter)
      FStrideSize: integer;  // step of kernel shift (in horizontal and also in vertical direction)
      FPaddingSize: integer; // padding of source feature map (left, top, right, bottom)
      Fforward_primitives_count,
      Fbackward_primitives_count: integer;
      Fforward_primitives,
      Fbackward_primitives: TCNNPrimitives;
      Fdst_memory_desc: dnnl_memory_desc_t;
      Fdst_memory,                      // memory used during forward propagation
      Fdiff_dst_memory: dnnl_memory_t;  // memory used during backward propagation
    protected
      procedure Initialize(const Asource: TCNNLayer; const AEngine: dnnl_engine_t); virtual;
      procedure ForwardPropagation; virtual;
      procedure BackwardPropagation; virtual;
    public
      constructor Create(ALayerType: TCNNLayerType; ADepth,AHeight,AWidth, AKernelSize,AStrideSize,APaddingSize: integer);
      destructor Destroy; override;
      function Data: PSingle; // feature map (forward propagation)
      function Diff: PSingle; // diff (backward propagation)
  end;

  TCNNLayerClass = class of TCNNLayer;

  { TCNNLayer__Convolution }

  TCNNLayer__Convolution = class(TCNNLayer)
    private
      Fic, Fih, Fiw: integer;
      Fweights_memory, Fbias_memory: dnnl_memory_t;
      Fdiff_weights_memory, Fdiff_bias_memory: dnnl_memory_t;
    protected
      procedure Initialize(const ASource: TCNNLayer; const AEngine: dnnl_engine_t); override;
      procedure BackwardPropagation; override;
    public
      destructor Destroy; override;
  end;

  { TCNNLayer_Convolution }

  TCNNLayer_Convolution = class(TCNNLayer__Convolution)
    protected
      procedure Initialize(const ASource: TCNNLayer; const AEngine: dnnl_engine_t); override;
  end;

  { TCNNLayer_FC }

  TCNNLayer_FC = class(TCNNLayer__Convolution)
    protected
      procedure Initialize(const ASource: TCNNLayer; const AEngine: dnnl_engine_t); override;
  end;

  { TCNNLayer_MaxPool }

  TCNNLayer_MaxPool = class(TCNNLayer)
    private
      Fws_memory: dnnl_memory_t;
    protected
      procedure Initialize(const ASource: TCNNLayer; const AEngine: dnnl_engine_t); override;
    public
      destructor Destroy; override;
  end;

  { TCNNLayer_AvgPool }

  TCNNLayer_AvgPool = class(TCNNLayer)
    private
    protected
      procedure Initialize(const ASource: TCNNLayer; const AEngine: dnnl_engine_t); override;
  end;

  { TCNNLayer__Eltwise }

  TCNNLayer__Eltwise = class(TCNNLayer)
    private
      Fforward_alg_kind, Fbackward_alg_kind: dnnl_alg_kind_t;
    protected
      procedure Initialize(const ASource: TCNNLayer; const AEngine: dnnl_engine_t); override;
  end;

  { TCNNLayer_ReLU }

  TCNNLayer_ReLU = class(TCNNLayer__Eltwise)
    protected
      procedure Initialize(const ASource: TCNNLayer; const AEngine: dnnl_engine_t); override;
  end;

  { TCNNLayer_LogSig }

  TCNNLayer_LogSig = class(TCNNLayer__Eltwise)
    protected
      procedure Initialize(const ASource: TCNNLayer; const AEngine: dnnl_engine_t); override;
  end;

  { TCNNLayer_Tanh }

  TCNNLayer_Tanh = class(TCNNLayer__Eltwise)
    protected
      procedure Initialize(const ASource: TCNNLayer; const AEngine: dnnl_engine_t); override;
  end;

  TCNNLayers = class(TList)
  end;

  ///////////////////////////////////////////////
  // simple/demo - Convolutional Neural Network
  ///////////////////////////////////////////////

  { TCNN }

  TCNN = class
    private
    var
      FEngine: dnnl_engine_t;
      FStream: dnnl_stream_t;
      FLayers: TCNNLayers;
      class procedure CHECK(status: dnnl_status_t);
    protected
    public
      constructor Create;
      destructor Destroy; override;
      // Add next convolutional layer
      function AddLayer(LayerType: TCNNLayerType; CountOfFeatureMaps, Height,Width, KernelSize,StrideSize,PaddingSize: integer): TCNNLayer;
      // Initialize memory structures etc.
      procedure Initialize;
      // Forward propagation
      function ForwardPropagation: integer;
      // Backward propagation
      procedure BackwardPropagation;
      // list of layers
      property Layers: TCNNLayers read FLayers;
  end;


{ TCNNLayer }

constructor TCNNLayer.Create(ALayerType: TCNNLayerType; ADepth, AHeight, AWidth,
  AKernelSize, AStrideSize, APaddingSize: integer);
begin
  inherited Create;
  FLayerType := ALayerType;
  FDepth := ADepth;
  FHeight := AHeight;
  FWidth := AWidth;
  FKernelSize := AKernelSize;
  FStrideSize := AStrideSize;
  FPaddingSize := APaddingSize;
end;

destructor TCNNLayer.Destroy;
var
  i: integer;
begin
  for i:=0 to Fforward_primitives_count-1 do begin
    if Assigned(Fforward_primitives[i].PrimitiveDesc) then
      dnnl_primitive_desc_destroy(Fforward_primitives[i].PrimitiveDesc);
    if Assigned(Fforward_primitives[i].Primitive) then
      dnnl_primitive_destroy(Fforward_primitives[i].Primitive);
  end;
  for i:=0 to Fbackward_primitives_count-1 do begin
    if Assigned(Fbackward_primitives[i].PrimitiveDesc) then
      dnnl_primitive_desc_destroy(FBackward_primitives[i].PrimitiveDesc);
    if Assigned(Fbackward_primitives[i].Primitive) then
      dnnl_primitive_destroy(Fbackward_primitives[i].Primitive);
  end;
  if Assigned(Fdst_memory) then
    dnnl_memory_destroy(Fdst_memory);
  if Assigned(Fdiff_dst_memory) then
    dnnl_memory_destroy(Fdiff_dst_memory);
  if Assigned(Fdst_memory_desc) then
    dnnl_memory_desc_destroy(Fdst_memory_desc);
  inherited;
end;

procedure TCNNLayer.Initialize(const Asource: TCNNLayer; const AEngine: dnnl_engine_t);
var
  dst_sizes, dst_padding_sizes, dst_offsets: dnnl_dims_t;
  tmp_memory_desc: dnnl_memory_desc_t;
  p: PSingle;
  i: integer;
begin
  // During the training phase, images are typically grouped together in batches.
  //   Even if there is only one image, we still assume that there is a batch with batch size equal to 1.
  //   Hence, the overall dimensionality of activations is 4D (N, C, H, and W)
  dst_sizes[0] := 1;
  dst_sizes[1] := FDepth;
  dst_sizes[2] := FHeight;
  dst_sizes[3] := FWidth;
  TCNN.CHECK(dnnl_memory_desc_create_with_tag(@Fdst_memory_desc, 4{Number of dimensions}, dst_sizes{Array of dimensions}, dnnl_f32, dnnl_nchw{dnnl_format_tag_any}));

  // create memory for dst data
  // The memory format and data type for src and dst are assumed to be the same. The same holds for diff_src and diff_dst.
  TCNN.CHECK(dnnl_memory_create(@Fdst_memory, Fdst_memory_desc, Aengine, DNNL_MEMORY_ALLOCATE));
  TCNN.CHECK(dnnl_memory_create(@Fdiff_dst_memory, Fdst_memory_desc, Aengine, DNNL_MEMORY_ALLOCATE));
  // Operation padding specifies how kernel will be applied to the source image, but it is logical, it doesn't require physically padded source image memory, implementations will adjust computations based on padding value.
  // Basically, padding is logical, so source memory is not padded physically in memory.
(*
  dst_padding_sizes[0] := 1;
  dst_padding_sizes[1] := FDepth;
  dst_padding_sizes[2] := FHeight+2*Fdst_padding;
  dst_padding_sizes[3] := FWidth+2*Fdst_padding;
  if Fdst_padding > 0 then begin
    // fill memory with 0
    TCNN.CHECK(dnnl_memory_get_data_handle(Fdst_memory, @p));
    for i:=0 to dnnl_memory_desc_get_size(@Fdst_memory_desc) div SizeOf(Single) - 1 do p[i]:=0;
    dst_sizes[0] := 1;
    dst_sizes[1] := FDepth;
    dst_sizes[2] := FHeight;
    dst_sizes[3] := FWidth;
    // dnnl_memory_desc_t.offset0
    dst_offsets[0] := 0;
    dst_offsets[1] := 0;
    dst_offsets[2] := Fdst_padding;
    dst_offsets[3] := Fdst_padding;
    tmp_memory_desc := Fdst_memory_desc;
    TCNN.CHECK(dnnl_memory_desc_init_submemory(@Fdst_memory_desc, @tmp_memory_desc, dst_sizes, dst_offsets));
  end;
*)
end;

procedure TCNNLayer.ForwardPropagation;
begin
end;

procedure TCNNLayer.BackwardPropagation;
begin
end;

function TCNNLayer.Data: PSingle;
begin
  TCNN.CHECK(dnnl_memory_get_data_handle(Fdst_memory, @Result));
end;

function TCNNLayer.Diff: PSingle;
begin
  TCNN.CHECK(dnnl_memory_get_data_handle(Fdiff_dst_memory, @Result));
end;

{ TCNNLayer__Convolution }

destructor TCNNLayer__Convolution.Destroy;
begin
  dnnl_memory_destroy(Fweights_memory);
  dnnl_memory_destroy(Fbias_memory);
  dnnl_memory_destroy(Fdiff_weights_memory);
  dnnl_memory_destroy(Fdiff_bias_memory);
  inherited;
end;

procedure TCNNLayer__Convolution.Initialize(const ASource: TCNNLayer; const AEngine: dnnl_engine_t);
const
  LEARNING_RATE=0.2;
var
  i: integer;
  conv_src_md, conv_weights_md, conv_bias_md: dnnl_memory_desc_t;
  conv_weights_sizes, conv_src_sizes: dnnl_dims_t; // 4D tensor
  conv_bias_sizes: dnnl_dims_t; // 1D tensor
  conv_strides, conv_dilates, conv_padding: dnnl_dims_t;
  scales: array[0..1] of Single;
  src_weights_update_md: array[0..1] of dnnl_memory_desc_t;
  src_bias_update_md: array[0..1] of dnnl_memory_desc_t;
  p: PSingle;
begin
  inherited;

  conv_src_sizes[0] := 1;
  conv_src_sizes[1] := Fic;
  conv_src_sizes[2] := Fih;
  conv_src_sizes[3] := Fiw;
  TCNN.CHECK(dnnl_memory_desc_create_with_tag(@conv_src_md, 4, conv_src_sizes, dnnl_f32, dnnl_nchw));

  conv_weights_sizes[0] := FDepth; // oc
  conv_weights_sizes[1] := Fic; // ic
  conv_weights_sizes[2] := FKernelSize; // kh
  conv_weights_sizes[3] := FKernelSize; // kw
  TCNN.CHECK(dnnl_memory_desc_create_with_tag(@conv_weights_md, 4, conv_weights_sizes, dnnl_f32, dnnl_oihw));

  conv_bias_sizes[0] := FDepth; // oc
  TCNN.CHECK(dnnl_memory_desc_create_with_tag(@conv_bias_md, 1, conv_bias_sizes, dnnl_f32, dnnl_x));

  conv_strides[0] := FStrideSize;
  conv_strides[1] := FStrideSize;
  conv_dilates[0] := 0;
  conv_dilates[1] := 0;
  conv_padding[0] := FPaddingSize;
  conv_padding[1] := FPaddingSize;
  // descriptor for a convolution forward propagation primitive
  TCNN.CHECK(dnnl_convolution_forward_primitive_desc_create(
          @Fforward_primitives[0].PrimitiveDesc, Aengine,
          dnnl_forward_training, dnnl_convolution_auto,
          conv_src_md, conv_weights_md, conv_bias_md, Fdst_memory_desc,
          conv_strides, conv_dilates, conv_padding, nil,
          nil));
  // descriptor for a convolution backward propagation primitive
  TCNN.CHECK(dnnl_convolution_backward_data_primitive_desc_create(
          @Fbackward_primitives[0].PrimitiveDesc, Aengine,
          dnnl_convolution_auto,
          conv_src_md, conv_weights_md, Fdst_memory_desc,
          conv_strides, conv_dilates, conv_padding, nil,
          Fforward_primitives[0].PrimitiveDesc,
          nil));
  // descriptor for a convolution weights gradient primitive
  TCNN.CHECK(dnnl_convolution_backward_weights_primitive_desc_create(
          @Fbackward_primitives[1].PrimitiveDesc, Aengine,
          dnnl_convolution_auto,
          conv_src_md, conv_weights_md, conv_bias_md, Fdst_memory_desc,
          conv_strides, conv_dilates, conv_padding, nil,
          Fforward_primitives[0].PrimitiveDesc,
          nil));

  // create weights memory
  TCNN.CHECK(dnnl_memory_create(@Fweights_memory, conv_weights_md, Aengine, DNNL_MEMORY_ALLOCATE));
  TCNN.CHECK(dnnl_memory_get_data_handle(Fweights_memory, @p));
  // fill weights memory
  for i:=0 to dnnl_memory_desc_get_size(conv_weights_md) div dnnl_data_type_size(dnnl_f32) - 1 do
    p[i]:={$IFDEF DEBUG_TCNN}1{$ELSE}(Random(){0-1}*2-1)/Sqrt(Fic*FKernelSize*FKernelSize){$ENDIF};

  // create bias memory
  TCNN.CHECK(dnnl_memory_create(@Fbias_memory, conv_bias_md, Aengine, DNNL_MEMORY_ALLOCATE));
  TCNN.CHECK(dnnl_memory_get_data_handle(Fbias_memory, @p));
  // fill bias memory
  for i:=0 to dnnl_memory_desc_get_size(conv_bias_md) div dnnl_data_type_size(dnnl_f32) - 1 do
    p[i]:=0;

  // create diff weights memory
  TCNN.CHECK(dnnl_memory_create(@Fdiff_weights_memory, conv_weights_md, Aengine, DNNL_MEMORY_ALLOCATE));
  // create diff bias memory
  TCNN.CHECK(dnnl_memory_create(@Fdiff_bias_memory, conv_bias_md, Aengine, DNNL_MEMORY_ALLOCATE));

  // forward propagation arguments
  Fforward_primitives_count := 1;
  Fforward_primitives[0].nargs := 4;
  Fforward_primitives[0].args[0].arg:=DNNL_ARG_SRC;     Fforward_primitives[0].args[0].memory := Asource.Fdst_memory;
  Fforward_primitives[0].args[1].arg:=DNNL_ARG_WEIGHTS; Fforward_primitives[0].args[1].memory := Fweights_memory;
  Fforward_primitives[0].args[2].arg:=DNNL_ARG_BIAS;    Fforward_primitives[0].args[2].memory := Fbias_memory;
  Fforward_primitives[0].args[3].arg:=DNNL_ARG_DST;     Fforward_primitives[0].args[3].memory := Fdst_memory;
  // backward propagation arguments
  Fbackward_primitives_count := 2;
  // The backward propagation computes diff_src based on diff_dst and weights
  Fbackward_primitives[0].nargs := 3;
  Fbackward_primitives[0].args[0].arg:=DNNL_ARG_DIFF_DST; Fbackward_primitives[0].args[0].memory := Fdiff_dst_memory;
  Fbackward_primitives[0].args[1].arg:=DNNL_ARG_WEIGHTS;  Fbackward_primitives[0].args[1].memory := Fweights_memory;
  Fbackward_primitives[0].args[2].arg:=DNNL_ARG_DIFF_SRC; Fbackward_primitives[0].args[2].memory := Asource.Fdiff_dst_memory;
  // The weights update computes diff_weights and diff_bias based on diff_dst and src
  Fbackward_primitives[1].nargs := 4;
  Fbackward_primitives[1].args[0].arg:=DNNL_ARG_DIFF_DST;     Fbackward_primitives[1].args[0].memory := Fdiff_dst_memory;
  Fbackward_primitives[1].args[1].arg:=DNNL_ARG_SRC;          Fbackward_primitives[1].args[1].memory := Asource.Fdst_memory;
  Fbackward_primitives[1].args[2].arg:=DNNL_ARG_DIFF_WEIGHTS; Fbackward_primitives[1].args[2].memory := Fdiff_weights_memory;
  Fbackward_primitives[1].args[3].arg:=DNNL_ARG_DIFF_BIAS;    Fbackward_primitives[1].args[3].memory := Fdiff_bias_memory;

  // Update weights and bias according to diff_weights and diff_bias
  scales[0] := 1;
  scales[1] := LEARNING_RATE;
  src_weights_update_md[0] := conv_weights_md;
  src_weights_update_md[1] := conv_weights_md;
  src_bias_update_md[0] := conv_bias_md;
  src_bias_update_md[1] := conv_bias_md;
  TCNN.CHECK(dnnl_sum_primitive_desc_create(@Fbackward_primitives[2].PrimitiveDesc, Aengine, conv_weights_md, 2, @scales[0], @src_weights_update_md[0], nil));
  TCNN.CHECK(dnnl_sum_primitive_desc_create(@Fbackward_primitives[3].PrimitiveDesc, Aengine, conv_bias_md   , 2, @scales[0], @src_bias_update_md[0]   , nil));
  Fbackward_primitives[2].nargs := 3;
  Fbackward_primitives[2].args[0].arg:=DNNL_ARG_MULTIPLE_SRC+1; Fbackward_primitives[2].args[0].memory := Fdiff_weights_memory;
  Fbackward_primitives[2].args[1].arg:=DNNL_ARG_MULTIPLE_SRC+0; Fbackward_primitives[2].args[1].memory := Fweights_memory;
  Fbackward_primitives[2].args[2].arg:=DNNL_ARG_DST;            Fbackward_primitives[2].args[2].memory := Fweights_memory;
  Fbackward_primitives[3].nargs := 3;
  Fbackward_primitives[3].args[0].arg:=DNNL_ARG_MULTIPLE_SRC+1; Fbackward_primitives[3].args[0].memory := Fdiff_bias_memory;
  Fbackward_primitives[3].args[1].arg:=DNNL_ARG_MULTIPLE_SRC+0; Fbackward_primitives[3].args[1].memory := Fbias_memory;
  Fbackward_primitives[3].args[2].arg:=DNNL_ARG_DST;            Fbackward_primitives[3].args[2].memory := Fbias_memory;

  dnnl_memory_desc_destroy(conv_src_md);
  dnnl_memory_desc_destroy(conv_weights_md);
  dnnl_memory_desc_destroy(conv_bias_md);
end;

procedure TCNNLayer__Convolution.BackwardPropagation;
const
  LEARNING_RATE=0.2;
var
  i: integer;
  weights, diff_weights, bias, diff_bias: PSingle;
  weights_md, bias_md: const_dnnl_memory_desc_t;
begin
(*
  // update weights
  dnnl_memory_get_data_handle(Fweights_memory, @weights);
  dnnl_memory_get_data_handle(Fdiff_weights_memory, @diff_weights);
  weights_md := dnnl_primitive_desc_query_md(Fforward_primitives[0].PrimitiveDesc, dnnl_query_weights_md, 0);
  for i:=0 to dnnl_memory_desc_get_size(weights_md) div SizeOf(Single) - 1 do
    weights[i] := weights[i] + diff_weights[i] * LEARNING_RATE;
  // update bias
  dnnl_memory_get_data_handle(Fbias_memory, @bias);
  dnnl_memory_get_data_handle(Fdiff_bias_memory, @diff_bias);
  bias_md := dnnl_primitive_desc_query_md(Fforward_primitives[0].PrimitiveDesc, dnnl_query_weights_md, 1);
  for i:=0 to dnnl_memory_desc_get_size(bias_md) div SizeOf(Single) - 1 do
    bias[i] := bias[i] + diff_bias[i] * LEARNING_RATE;
*)
end;

{ TCNNLayer_Convolution }

procedure TCNNLayer_Convolution.Initialize(const Asource: TCNNLayer; const AEngine: dnnl_engine_t);
begin
  Fic := ASource.FDepth;
  Fih := ASource.FHeight;
  Fiw := ASource.FWidth;
  inherited;
end;

{ TCNNLayer_FC }

procedure TCNNLayer_FC.Initialize(const ASource: TCNNLayer; const AEngine: dnnl_engine_t);
begin
  Fic := ASource.FDepth*ASource.FHeight*ASource.FWidth; // input channels
  Fih := 1;
  Fiw := 1;
  FHeight := 1;
  FWidth := 1;
  FKernelSize := 1;
  FStrideSize := 1;
  FPaddingSize := 0;
  inherited;
end;

{ TCNNLayer_MaxPool }

destructor TCNNLayer_MaxPool.Destroy;
begin
  dnnl_memory_destroy(Fws_memory);
  inherited;
end;

procedure TCNNLayer_MaxPool.Initialize(const Asource: TCNNLayer; const AEngine: dnnl_engine_t);
var
  pool_strides, pool_kernel, pool_dilation, pool_padding_l, pool_padding_r: dnnl_dims_t;
  pool_ws_md: const_dnnl_memory_desc_t;
begin
  inherited;

  pool_strides[0] := FStrideSize;
  pool_strides[1] := FStrideSize;
  pool_kernel[0] := FKernelSize;
  pool_kernel[1] := FKernelSize;
  pool_dilation[0] := 0;
  pool_dilation[1] := 0;
  pool_padding_l[0] := FPaddingSize; // PH_l
  pool_padding_l[1] := FPaddingSize; // PW_l
  pool_padding_r[0] := (FHeight-1)*FStrideSize+FKernelSize-Asource.FHeight-FPaddingSize; // PH_r
  pool_padding_r[1] := (FWidth -1)*FStrideSize+FKernelSize-Asource.FWidth -FPaddingSize; // PW_r
  // descriptor for pooling forward propagation primitive
  TCNN.CHECK(dnnl_pooling_forward_primitive_desc_create(
          @Fforward_primitives[0].PrimitiveDesc, Aengine,
          dnnl_forward_training, dnnl_pooling_max,
          Asource.Fdst_memory_desc, Fdst_memory_desc,
          pool_strides, pool_kernel, pool_dilation, pool_padding_l, pool_padding_r,
          nil));
  // descriptor for pooling backward propagation primitive
  TCNN.CHECK(dnnl_pooling_backward_primitive_desc_create(
          @Fbackward_primitives[0].PrimitiveDesc, Aengine,
          dnnl_pooling_max,
          Asource.Fdst_memory_desc, Fdst_memory_desc,
          pool_strides, pool_kernel, pool_dilation, pool_padding_l, pool_padding_r,
          Fforward_primitives[0].PrimitiveDesc,
          nil));

  // create memory for workspace
  // (Workspace tensor that carries additional information from the forward propagation to the backward propagation)
  pool_ws_md := dnnl_primitive_desc_query_md(Fforward_primitives[0].PrimitiveDesc, dnnl_query_workspace_md, 0);
  dnnl_memory_create(@Fws_memory, pool_ws_md, Aengine, DNNL_MEMORY_ALLOCATE);

  // forward propagation arguments
  Fforward_primitives_count := 1;
  Fforward_primitives[0].nargs := 3;
  Fforward_primitives[0].args[0].arg:=DNNL_ARG_SRC;       Fforward_primitives[0].args[0].memory := Asource.Fdst_memory;
  Fforward_primitives[0].args[1].arg:=DNNL_ARG_DST;       Fforward_primitives[0].args[1].memory := Fdst_memory;
  Fforward_primitives[0].args[2].arg:=DNNL_ARG_WORKSPACE; Fforward_primitives[0].args[2].memory := Fws_memory;
  // backward propagation arguments
  // (The backward propagation computes diff_src, based on diff_dst and (in case of max pooling) workspace)
  Fbackward_primitives_count := 1;
  Fbackward_primitives[0].nargs := 3;
  Fbackward_primitives[0].args[0].arg:=DNNL_ARG_DIFF_DST;  Fbackward_primitives[0].args[0].memory := Fdiff_dst_memory;
  Fbackward_primitives[0].args[1].arg:=DNNL_ARG_DIFF_SRC;  Fbackward_primitives[0].args[1].memory := Asource.Fdiff_dst_memory;
  Fbackward_primitives[0].args[2].arg:=DNNL_ARG_WORKSPACE; Fbackward_primitives[0].args[2].memory := Fws_memory;
end;

{ TCNNLayer_AvgPool }

procedure TCNNLayer_AvgPool.Initialize(const Asource: TCNNLayer; const AEngine: dnnl_engine_t);
var
  pool_strides, pool_kernel, pool_dilation, pool_padding_l, pool_padding_r: dnnl_dims_t;
begin
  inherited;

  pool_strides[0] := FStrideSize;
  pool_strides[1] := FStrideSize;
  pool_kernel[0] := FKernelSize;
  pool_kernel[1] := FKernelSize;
  pool_dilation[0] := 0;
  pool_dilation[1] := 0;
  pool_padding_l[0] := FPaddingSize; // PH_l
  pool_padding_l[1] := FPaddingSize; // PW_l
  pool_padding_r[0] := (FHeight-1)*FStrideSize+FKernelSize-Asource.FHeight-FPaddingSize; // PH_r
  pool_padding_r[1] := (FWidth -1)*FStrideSize+FKernelSize-Asource.FWidth -FPaddingSize; // PW_r
  // descriptor for pooling forward propagation primitive
  TCNN.CHECK(dnnl_pooling_forward_primitive_desc_create(
          @Fforward_primitives[0].PrimitiveDesc, Aengine,
          dnnl_forward_training, dnnl_pooling_avg_exclude_padding,
          Asource.Fdst_memory_desc, Fdst_memory_desc,
          pool_strides, pool_kernel, pool_dilation, pool_padding_l, pool_padding_r,
          nil));
  // descriptor for pooling backward propagation primitive
  TCNN.CHECK(dnnl_pooling_backward_primitive_desc_create(
          @Fbackward_primitives[0].PrimitiveDesc, Aengine,
          dnnl_pooling_avg_exclude_padding,
          Asource.Fdst_memory_desc, Fdst_memory_desc,
          pool_strides, pool_kernel, pool_dilation, pool_padding_l, pool_padding_r,
          Fforward_primitives[0].PrimitiveDesc,
          nil));

  // forward propagation arguments
  Fforward_primitives_count := 1;
  Fforward_primitives[0].nargs := 2;
  Fforward_primitives[0].args[0].arg:=DNNL_ARG_SRC; Fforward_primitives[0].args[0].memory := Asource.Fdst_memory;
  Fforward_primitives[0].args[1].arg:=DNNL_ARG_DST; Fforward_primitives[0].args[1].memory := Fdst_memory;
  // backward propagation arguments
  Fbackward_primitives_count := 1;
  Fbackward_primitives[0].nargs := 2;
  Fbackward_primitives[0].args[0].arg:=DNNL_ARG_DIFF_DST; Fbackward_primitives[0].args[0].memory := Fdiff_dst_memory;
  Fbackward_primitives[0].args[1].arg:=DNNL_ARG_DIFF_SRC; Fbackward_primitives[0].args[1].memory := Asource.Fdiff_dst_memory;
end;

{ TCNNLayer__Eltwise }

procedure TCNNLayer__Eltwise.Initialize(const Asource: TCNNLayer; const AEngine: dnnl_engine_t);
begin
  inherited;

  // There is no difference between the dnnl_forward_training and dnnl_forward_inference propagation kinds
  TCNN.CHECK(dnnl_eltwise_forward_primitive_desc_create(
          @Fforward_primitives[0].PrimitiveDesc, Aengine,
          dnnl_forward_training, Fforward_alg_kind,
          Asource.Fdst_memory_desc, Fdst_memory_desc, 0{negative_slope}, 0,
          nil));
  TCNN.CHECK(dnnl_eltwise_backward_primitive_desc_create(
          @Fbackward_primitives[0].PrimitiveDesc, Aengine,
          Fbackward_alg_kind,
          Asource.Fdst_memory_desc, Fdst_memory_desc, Fdst_memory_desc, 0{negative_slope}, 0,
          Fforward_primitives[0].PrimitiveDesc,
          nil));

  // forward propagation arguments
  Fforward_primitives_count := 1;
  Fforward_primitives[0].nargs := 2;
  Fforward_primitives[0].args[0].arg:=DNNL_ARG_SRC; Fforward_primitives[0].args[0].memory := Asource.Fdst_memory;
  Fforward_primitives[0].args[1].arg:=DNNL_ARG_DST; Fforward_primitives[0].args[1].memory := Fdst_memory;
  // The backward propagation computes "diff_src" based on "diff_dst" and "src" tensors.
  // However, some operations support a computation using "dst" memory produced during the forward propagation.
  Fbackward_primitives_count := 1;
  Fbackward_primitives[0].nargs := 3;
  Fbackward_primitives[0].args[0].arg:=DNNL_ARG_DST;      Fbackward_primitives[0].args[0].memory := Fdst_memory;
  Fbackward_primitives[0].args[1].arg:=DNNL_ARG_DIFF_DST; Fbackward_primitives[0].args[1].memory := Fdiff_dst_memory;
  Fbackward_primitives[0].args[2].arg:=DNNL_ARG_DIFF_SRC; Fbackward_primitives[0].args[2].memory := Asource.Fdiff_dst_memory;
end;

{ TCNNLayer_ReLU }

procedure TCNNLayer_ReLU.Initialize(const Asource: TCNNLayer; const AEngine: dnnl_engine_t);
begin
  Fforward_alg_kind := dnnl_eltwise_relu;
  Fbackward_alg_kind := dnnl_eltwise_relu_use_dst_for_bwd;
  inherited;
end;

{ TCNNLayer_LogSig }

procedure TCNNLayer_LogSig.Initialize(const Asource: TCNNLayer; const AEngine: dnnl_engine_t);
begin
  Fforward_alg_kind := dnnl_eltwise_logistic;
  Fbackward_alg_kind := dnnl_eltwise_logistic_use_dst_for_bwd;
  inherited;
end;

{ TCNNLayer_Tanh }

procedure TCNNLayer_Tanh.Initialize(const Asource: TCNNLayer; const AEngine: dnnl_engine_t);
begin
  Fforward_alg_kind := dnnl_eltwise_tanh;
  Fbackward_alg_kind := dnnl_eltwise_tanh_use_dst_for_bwd;
  inherited;
end;

{ TCNN }

constructor TCNN.Create;
begin
  inherited Create;
  // create layers
  FLayers := TCNNLayers.Create;
end;

destructor TCNN.Destroy;
var
  i: integer;
begin
  // destroy stream
  CHECK(dnnl_stream_destroy(Fstream));
  // destroy engine
  CHECK(dnnl_engine_destroy(Fengine));

  // destroy layers
  for i:=FLayers.Count-1 downto 0 do begin
    TCNNLayers(FLayers[i]).Free;
  end;
  FLayers.Free;

  inherited;
end;

class procedure TCNN.CHECK(status: dnnl_status_t);
begin
  if status <> dnnl_success then raise Exception.CreateFmt('DNNL Error : %d', [ord(status)]);
end;

function TCNN.AddLayer(LayerType: TCNNLayerType; CountOfFeatureMaps,
  Height, Width, KernelSize, StrideSize, PaddingSize: integer): TCNNLayer;
var
  layerClass: TCNNLayerClass;
begin
  case LayerType of
    cltInput:          layerClass := TCNNLayer;
    cltConvolution:    layerClass := TCNNLayer_Convolution;
    cltFuncReLU:       layerClass := TCNNLayer_ReLU;
    cltMaxPooling:     layerClass := TCNNLayer_MaxPool;
    cltAvgPooling:     layerClass := TCNNLayer_AvgPool;
    cltFullyConnected: layerClass := TCNNLayer_FC;
    cltFuncLogSig:     layerClass := TCNNLayer_LogSig;
    cltFuncTanh:       layerClass := TCNNLayer_Tanh;
  end;
  Result := layerClass.Create(LayerType, CountOfFeatureMaps, Height, Width, KernelSize, StrideSize, PaddingSize);
  FLayers.Add(Result);
end;

procedure TCNN.Initialize;
var
  i, j: integer;
  srcLayer, aLayer: TCNNLayer;
begin
  // Engine is abstraction of a computational device: a CPU, a specific GPU card in the system, etc.
  // Most primitives are created to execute computations on one specific engine.
  CHECK(dnnl_engine_create(@Fengine, dnnl_cpu, 0));

  // The primitive’s computations are executed on the computational device corresponding to the engine on which the primitive (and memory arguments) were created and happens within the context of the stream.
  CHECK(dnnl_stream_create(@Fstream, Fengine, ord(dnnl_stream_default_flags)));

  // Propagate paddings from the last layer to the first layer
(*
  padding := 0;
  for i:=FLayers.Count-1 downto 0 do begin
    TCNNLayer(FLayers[i]).Fdst_padding := padding;
    padding := TCNNLayer(FLayers[i]).FPaddingSize;
  end;
*)
  // Create forward and backward primitives from layers primitive descriptors
  srcLayer := TCNNLayer(FLayers[0]);
  srcLayer.Initialize(nil, Fengine);
  for i:=1 to FLayers.Count-1 do begin
    aLayer := TCNNLayer(FLayers[i]);
    // initializes memory structures, primitive descriptors, arguments for individual primitives
    aLayer.Initialize(srcLayer, Fengine);

    // create forward primitives
    for j:=0 to aLayer.Fforward_primitives_count-1 do
      CHECK(dnnl_primitive_create(@aLayer.Fforward_primitives[j].Primitive , aLayer.Fforward_primitives[j].PrimitiveDesc));

    // create backward primitive
    for j:=0 to aLayer.Fbackward_primitives_count-1 do
      CHECK(dnnl_primitive_create(@aLayer.Fbackward_primitives[j].Primitive, aLayer.Fbackward_primitives[j].PrimitiveDesc));

    srcLayer := aLayer;
  end;
end;

function TCNN.ForwardPropagation: integer;
var
  i, j: integer;
  primitive: TCNNPrimitive;
  status: dnnl_status_t;
begin
  for i:=0 to FLayers.Count-1 do begin
    // execution of individual forward primitives
    for j:=0 to TCNNLayer(FLayers[i]).Fforward_primitives_count-1 do begin
      primitive := TCNNLayer(FLayers[i]).Fforward_primitives[j];
      status := dnnl_primitive_execute(primitive.Primitive, Fstream, primitive.nargs, primitive.args);
      if status <> dnnl_success then
         raise Exception.CreateFmt('Error %d executing DNNL forward primitive #%d',[status, i]);
    end;
  end;
  CHECK(dnnl_stream_wait(Fstream));
end;

procedure TCNN.BackwardPropagation;
var
  i,j: integer;
  primitive: TCNNPrimitive;
  status: dnnl_status_t;
begin
  // execution of individual backward primitives
  for i:=FLayers.Count-1 downto 0 do begin
    // execution of individual backward primitives
    for j:=0 to TCNNLayer(FLayers[i]).Fbackward_primitives_count-1 do begin
      primitive := TCNNLayer(FLayers[i]).Fbackward_primitives[j];
      status := dnnl_primitive_execute(primitive.Primitive, Fstream, primitive.nargs, primitive.args);
      if status <> dnnl_success then
         raise Exception.CreateFmt('Error %d executing DNNL backward primitive #%d',[status, i]);
    end;
    TCNNLayer(FLayers[i]).BackwardPropagation;
  end;
  CHECK(dnnl_stream_wait(Fstream));
end;

{$IFNDEF HAVE_TCNN}
procedure CHECK(status: dnnl_status_t);
begin
  if status <> dnnl_success then WriteLn('Error : ', status);
end;

const
  // IC-InputChannel, OC-OutputChannel, H-Height, W-Width, KH-KernelHeight, KW-KernelWidth
  zero_padding: dnnl_dims_t = (0,0,0,0,0,0,0,0,0,0,0,0);
  zero_dilates: dnnl_dims_t = (0,0,0,0,0,0,0,0,0,0,0,0);

  // 6x6x1 -> 6x6x8, 8 Feature maps
  conv_src_offsets: dnnl_dims_t = (0,0,1,1,0,0,0,0,0,0,0,0);
  conv_src_sizes: dnnl_dims_t = (1{n},1{c},6{h},6{w},0,0,0,0,0,0,0,0); // 4D tensor
  conv_weights_sizes: dnnl_dims_t = (8{oc},1{ic},3{kh},3{kw},0,0,0,0,0,0,0,0); // 4D tensor
  conv_bias_sizes: dnnl_dims_t = (8{oc},0,0,0,0,0,0,0,0,0,0,0); // 1D tensor
  conv_dst_sizes: dnnl_dims_t = (1{n},8{oc},6{h},6{w},0,0,0,0,0,0,0,0); // 4D tensor
  conv_strides: dnnl_dims_t = (1,1,0,0,0,0,0,0,0,0,0,0);
  conv_padding: dnnl_dims_t = (1,1,0,0,0,0,0,0,0,0,0,0);

  // 6x6x8 -> 3x3x8
  pool_dst_sizes: dnnl_dims_t = (1{n},8{c},3{h},3{w},0,0,0,0,0,0,0,0);
  pool_strides: dnnl_dims_t = (2,2,0,0,0,0,0,0,0,0,0,0);
  pool_kernel: dnnl_dims_t = (2,2,0,0,0,0,0,0,0,0,0,0);

  // 3x3x8 -> 1x1x10
  fc_src_sizes: dnnl_dims_t = (1{n},3*3*8{c},1{h},1{w},0,0,0,0,0,0,0,0);
  fc_weights_sizes: dnnl_dims_t = (10{oc},3*3*8{ic},1{kh},1{kw},0,0,0,0,0,0,0,0); // 4D tensor
  fc_bias_sizes: dnnl_dims_t = (10{oc},0,0,0,0,0,0,0,0,0,0,0); // 1D tensor
  fc_dst_sizes: dnnl_dims_t = (1{n},10{oc},1{h},1{w},0,0,0,0,0,0,0,0); // 4D tensor
  fc_strides: dnnl_dims_t = (1,1,0,0,0,0,0,0,0,0,0,0);

type
  TPrimitive = record
    primitive: dnnl_primitive_t;
    nargs: int;
    args: Pdnnl_exec_arg_t;
  end;
{$ENDIF}

var
  i: integer;
  p: PSingle;

  version: Pdnnl_version_t;

{$IFDEF HAVE_TCNN}
  cnn: TCNN;
{$ELSE}
  engine: dnnl_engine_t;
  // simple net
  // fowrard
  net_fwd: array[0..3] of TPrimitive;
  net_f: integer;
  // backward:
  net_bwd: array[0..3] of TPrimitive;
  net_b: integer;

  input_src_md: dnnl_memory_desc_t;
  input_src_memory: dnnl_memory_t;

  conv_src_md: dnnl_memory_desc_t;
  conv_src_memory: dnnl_memory_t;
  conv_weights_md, conv_bias_md: dnnl_memory_desc_t;
  conv_dst_md: const_dnnl_memory_desc_t;
  conv_weights_memory, conv_bias_memory, conv_dst_memory: dnnl_memory_t;
  conv_pd: dnnl_primitive_desc_t;
  conv: dnnl_primitive_t;
  conv_args: array[0..3] of dnnl_exec_arg_t;

  relu_src_md: const_dnnl_memory_desc_t;
  relu_pd: dnnl_primitive_desc_t;
  relu_dst_md: const_dnnl_memory_desc_t;
  relu_dst_memory, relu_src_memory: dnnl_memory_t;
  relu: dnnl_primitive_t;
  relu_args: array[0..1] of dnnl_exec_arg_t;

  // create descriptors for src, dst pooling data
  pool_src_md: const_dnnl_memory_desc_t;
  pool_dst_md: const_dnnl_memory_desc_t;
  pool_ws_md: const_dnnl_memory_desc_t;
  pool_dst_memory, pool_ws_memory: dnnl_memory_t;
  // create a pooling primitive descriptor
  pool_pd: dnnl_primitive_desc_t;
  pool: dnnl_primitive_t;
  pool_args: array[0..2] of dnnl_exec_arg_t;

  fc_src_md: dnnl_memory_desc_t;
  fc_weights_md, fc_bias_md: dnnl_memory_desc_t;
  fc_dst_md: dnnl_memory_desc_t;
  fc_weights_memory, fc_bias_memory, fc_dst_memory: dnnl_memory_t;
  fc_pd: dnnl_primitive_desc_t;
  fc: dnnl_primitive_t;
  fc_args: array[0..3] of dnnl_exec_arg_t;

  pool_diff_src_md, pool_diff_dst_md: const_dnnl_memory_desc_t;
  pool_diff_src_memory, pool_diff_dst_memory: dnnl_memory_t;
  pool_bwd_pd: dnnl_primitive_desc_t;
  pool_bwd: dnnl_primitive_t;
  pool_bwd_args: array[0..2] of dnnl_exec_arg_t;

  relu_diff_src_memory: dnnl_memory_t;
  relu_bwd_pd: dnnl_primitive_desc_t;
  relu_bwd: dnnl_primitive_t;
  relu_bwd_args: array[0..2] of dnnl_exec_arg_t;

  stream: dnnl_stream_t;
{$ENDIF}
begin
  version := dnnl_version();
  WriteLn('DNNL version: ', version^.major, '.', version^.minor, '.', version^.patch);
{$IFDEF HAVE_TCNN}
  // instance of CNN
  cnn := TCNN.Create;
  // #0: add input layer (gray-scale image): 1x6x6
  cnn.AddLayer(cltInput, 1{channel}, W, H, 0,0,0);
  // #1: add 1. convolution layer: 8x6x6
  cnn.AddLayer(cltConvolution, 8{depth}, 6{width}, 6{height}, 3{kernel}, 1{stride}, 1{padding});
  // #2: add ReLU layer
  cnn.AddLayer(cltFuncReLU, 8{depth}, 6{width}, 6{height}, 0,0,0);
  // #3: add 1. MaxPool layer: 8x3x3
  cnn.AddLayer(cltMaxPooling, 8{depth}, 3{width}, 3{height}, 2{kernel}, 2{stride}, 0);
  // #4: add 1. fully connected layer
  cnn.AddLayer(cltFullyConnected, 10{depth/neurons}, 1{width}, 1{height}, 1{kernel}, 1{stride}, 0{padding});

  // create structures for forward and backward propagation
  cnn.Initialize;

  p := TCNNLayer(cnn.Layers[0]).Data;
  for i:=Low(input_) to High(input_) do p[i]:=input_[i];

  cnn.ForwardPropagation;

  p := TCNNLayer(cnn.Layers[4]).Data;
  for i:=0 to 10-1 do Write(p[i]:6:0, ' ');

  // here we can use cnn.Diff to set deltas for backward propagation
  p := TCNNLayer(cnn.Layers[4]).Diff;
  for i:=0 to 10-1 do p[i]:=0;
  p[0] := 1;

  cnn.BackwardPropagation;

  cnn.Free;
{$ELSE}
  // Engine is abstraction of a computational device: a CPU, a specific GPU card in the system, etc.
  // Most primitives are created to execute computations on one specific engine.
  CHECK(dnnl_engine_create(@engine, dnnl_cpu, 0));

  {----------------------------------------------------------------------------}
  {--------------------------- Forward Stream ---------------------------------}
  {----------------------------------------------------------------------------}
  net_f := 0;

  (*****************************)
  (*** INPUT GRAYSCALE IMAGE ***)
  (*****************************)
  Writeln('Input image');
  CHECK(dnnl_memory_desc_create_with_tag(@input_src_md, 4{Number of dimensions}, input_src_sizes{Array of dimensions}, dnnl_f32, dnnl_nchw));
  Writeln('  memory size: ',dnnl_memory_desc_get_size(input_src_md)); // W*H*4

  // create memory for input src data
  CHECK(dnnl_memory_create(@input_src_memory, input_src_md, engine, DNNL_MEMORY_ALLOCATE));
  CHECK(dnnl_memory_get_data_handle(input_src_memory, @p));
  for i:=Low(input_) to High(input_) do p[i]:=input_[i];

  (*************************)
  (*** CONVOLUTION LAYER ***)
  (*************************)
  Writeln('Convolution');
  conv_src_md := input_src_md;
  //CHECK(dnnl_memory_desc_init_submemory(@conv_src_md, @input_src_md, conv_src_sizes, conv_src_offsets));
  // Submemory is created with a pointer to the parent buffer. As a result submemory doesn't own allocated memory, but allows to interpret part of it as a separate memory entity.
  // Size is equal 0, because it doesn't own the buffer (buffer belongs to parent memory and will be destroyed as part of parent memory).
  Writeln('  memory size: ',dnnl_memory_desc_get_size(conv_src_md)); // 0
  conv_src_memory := input_src_memory;

  // During the training phase, images are typically grouped together in batches.
  //   Even if there is only one image, we still assume that there is a batch with batch size equal to 1.
  //   Hence, the overall dimensionality of activations is 4D (N, C, H, and W)
  // create data descriptors for convolution w/ no specified format
  // the convolution primitive expects the following tensors:
  //   2D: Source/Destination: N x C x H x W;
  //                  Weights: OC x IC x KH x KW
  CHECK(dnnl_memory_desc_create_with_tag(@conv_weights_md, 4, conv_weights_sizes, dnnl_f32, dnnl_oihw));
  CHECK(dnnl_memory_desc_create_with_tag(@conv_bias_md, 1, conv_bias_sizes, dnnl_f32, dnnl_x));
  CHECK(dnnl_memory_desc_create_with_tag(@conv_dst_md, 4, conv_dst_sizes, dnnl_f32, dnnl_format_tag_any));

  // Operation padding specifies how kernel will be applied to the source image, but it is logical, it doesn't require physically padded source image memory, implementations will adjust computations based on padding value.
  // Basically, padding is logical, so source memory is not padded physically in memory.
  CHECK(dnnl_convolution_forward_primitive_desc_create(
          @conv_pd, engine,
          dnnl_forward, dnnl_convolution_direct,
          conv_src_md, conv_weights_md, conv_bias_md, conv_dst_md,
          conv_strides, zero_dilates, conv_padding, conv_padding,
          nil));

  // create weights memory
  //conv_weights_md := dnnl_primitive_desc_query_md(conv_pd, dnnl_query_weights_md, 0);
  CHECK(dnnl_memory_create(@conv_weights_memory, conv_weights_md, engine, DNNL_MEMORY_ALLOCATE));
  CHECK(dnnl_memory_get_data_handle(conv_weights_memory, @p));
  for i:=0 to dnnl_memory_desc_get_size(conv_weights_md) div 4 - 1 do p[i]:=1;

  // create bias memory
  CHECK(dnnl_memory_create(@conv_bias_memory, conv_bias_md, engine, DNNL_MEMORY_ALLOCATE));
  CHECK(dnnl_memory_get_data_handle(conv_bias_memory, @p));
  for i:=0 to dnnl_memory_desc_get_size(conv_bias_md) div 4 - 1 do p[i]:=0;

  // create conv dst memory
  // When using any it is necessary to first create a convolution primitive descriptor and then query it for the actual data and weight memory objects formats.
  conv_dst_md := dnnl_primitive_desc_query_md(conv_pd, dnnl_query_dst_md, 0);
  CHECK(dnnl_memory_create(@conv_dst_memory, conv_dst_md, engine, DNNL_MEMORY_ALLOCATE));

  // finally create a convolution primitive
  CHECK(dnnl_primitive_create(@conv, conv_pd));
  net_fwd[net_f].primitive := conv;
  net_fwd[net_f].nargs:=4;
  net_fwd[net_f].args:=@conv_args[0];
  conv_args[0].arg:=DNNL_ARG_SRC;     conv_args[0].memory := conv_src_memory;
  conv_args[1].arg:=DNNL_ARG_WEIGHTS; conv_args[1].memory := conv_weights_memory;
  conv_args[2].arg:=DNNL_ARG_BIAS;    conv_args[2].memory := conv_bias_memory;
  conv_args[3].arg:=DNNL_ARG_DST;     conv_args[3].memory := conv_dst_memory;
  Inc(net_f);

  (************)
  (*** ReLU ***)
  (************)
  // element wise operation: x<=0 => 0, x>0 => x
  Writeln('ReLU');

  // keep memory format of source same as the format of convolution
  // output in order to avoid reorder
  relu_src_md := conv_dst_md;
  relu_src_memory := conv_dst_memory;

  // create a relu primitive descriptor
  CHECK(dnnl_eltwise_forward_primitive_desc_create(
          @relu_pd, engine,
          dnnl_forward, dnnl_eltwise_relu,
          relu_src_md, relu_src_md,
          0{negative_slope}, 0,
          nil));

  // create relu dst memory
  relu_dst_md := dnnl_primitive_desc_query_md(relu_pd, dnnl_query_dst_md, 0);
  CHECK(dnnl_memory_create(@relu_dst_memory, relu_dst_md, engine, DNNL_MEMORY_ALLOCATE));

  // finally create a relu primitive
  CHECK(dnnl_primitive_create(@relu, relu_pd));
  net_fwd[net_f].primitive := relu;
  net_fwd[net_f].nargs:=2;
  net_fwd[net_f].args:=@relu_args[0];
  relu_args[0].arg:=DNNL_ARG_SRC; relu_args[0].memory := relu_src_memory;
  relu_args[1].arg:=DNNL_ARG_DST; relu_args[1].memory := relu_dst_memory;
  Inc(net_f);

  (*************************)
  (*** MAX-POOLING LAYER ***)
  (*************************)
  writeln('Max Pooling');
  // N x C x H x W tensor for the 2D spatial case (4 dimensions)
  //   c-channels, d-depth, h-height, w-width
  //   (c-channels are the same as feature maps)
  //   - offset_nchw(n, c, h, w) = n * CHW + c * HW + h * W + w;
  //     data are kept packed or in dense form, meaning pixels follow one another
  //   - offset(n, c, h, w) = n * stride_n + c * stride_c + h * stride_h + w * stride_w;
  //     in case of dnnl_nchw format: stride_n = CHW, stride_c = HW, stride_h = W, stride_w = 1
  //     In particular, whenever a user creates memory with the dnnl_nchw format, DNNL computes the strides and fills the structure on behalf of the user.
  // One can create memory with NCHW data layout using dnnl_nchw
  pool_src_md := relu_dst_md;
  CHECK(dnnl_memory_desc_create_with_tag(@pool_dst_md, 4{Number of dimensions}, pool_dst_sizes{Array of dimensions}, dnnl_f32, dnnl_nchw{dnnl_format_tag_any}));

  // Arrays strides, kernel, padding_l, and padding_r contain values for spatial dimensions only
  //  and hence must have the same number of elements as there are spatial dimensions.
  CHECK(dnnl_pooling_forward_primitive_desc_create(
          @pool_pd, engine,
          dnnl_forward_training, dnnl_pooling_max,
          pool_src_md, pool_dst_md,
          @pool_strides[0], @pool_kernel[0], @zero_dilates[0], @zero_padding[0], nil,
          nil));

  // create memory for dst data
  CHECK(dnnl_memory_create(@pool_dst_memory, pool_dst_md, engine, DNNL_MEMORY_ALLOCATE));

  // create memory for workspace
  // (Workspace tensor that carries additional information from the forward propagation to the backward propagation)
  pool_ws_md := dnnl_primitive_desc_query_md(pool_pd, dnnl_query_workspace_md, 0);
  CHECK(dnnl_memory_create(@pool_ws_memory, pool_ws_md, engine, DNNL_MEMORY_ALLOCATE));

  // finally create a pooling primitive
  CHECK(dnnl_primitive_create(@pool, pool_pd));
  net_fwd[net_f].primitive := pool;

  // pool operation has 3 arguments for execution:
  net_fwd[net_f].nargs:=3;
  net_fwd[net_f].args:=@pool_args[0];
  pool_args[0].arg:=DNNL_ARG_SRC;       pool_args[0].memory := relu_dst_memory;
  pool_args[1].arg:=DNNL_ARG_DST;       pool_args[1].memory := pool_dst_memory;
  pool_args[2].arg:=DNNL_ARG_WORKSPACE; pool_args[2].memory := pool_ws_memory;
  Inc(net_f);

  (*****************************)
  (*** FULLY-CONNECTED LAYER ***)
  (*****************************)
  Writeln('Fully connected');

  CHECK(dnnl_memory_desc_create_with_tag(@fc_src_md, 4, fc_src_sizes, dnnl_f32, dnnl_nchw));
  CHECK(dnnl_memory_desc_create_with_tag(@fc_weights_md, 4, fc_weights_sizes, dnnl_f32, dnnl_oihw));
  CHECK(dnnl_memory_desc_create_with_tag(@fc_bias_md, 1, fc_bias_sizes, dnnl_f32, dnnl_x));
  CHECK(dnnl_memory_desc_create_with_tag(@fc_dst_md, 4, fc_dst_sizes, dnnl_f32, dnnl_nchw{dnnl_format_tag_any}));

  CHECK(dnnl_convolution_forward_primitive_desc_create(
          @fc_pd, engine,
          dnnl_forward_training, dnnl_convolution_direct,
          fc_src_md, fc_weights_md, fc_bias_md, fc_dst_md,
          fc_strides, zero_dilates, zero_padding, nil,
          nil));

  // create weights memory
  CHECK(dnnl_memory_create(@fc_weights_memory, fc_weights_md, engine, DNNL_MEMORY_ALLOCATE));
  CHECK(dnnl_memory_get_data_handle(fc_weights_memory, @p));
  for i:=0 to dnnl_memory_desc_get_size(fc_weights_md) div SizeOf(Single) - 1 do p[i]:=1;

  // create bias memory
  CHECK(dnnl_memory_create(@fc_bias_memory, fc_bias_md, engine, DNNL_MEMORY_ALLOCATE));
  CHECK(dnnl_memory_get_data_handle(fc_bias_memory, @p));
  for i:=0 to dnnl_memory_desc_get_size(fc_bias_md) div SizeOf(Single) - 1 do p[i]:=0;

  // create dst memory
  // When using any it is necessary to first create a convolution primitive descriptor and then query it for the actual data and weight memory objects formats.
  //fc_dst_md := dnnl_primitive_desc_query_md(fc_pd, dnnl_query_dst_md, 0);
  CHECK(dnnl_memory_create(@fc_dst_memory, fc_dst_md, engine, DNNL_MEMORY_ALLOCATE));

  // finally create a convolution primitive
  CHECK(dnnl_primitive_create(@fc, fc_pd));
  net_fwd[net_f].primitive := fc;
  net_fwd[net_f].nargs:=4;
  net_fwd[net_f].args:=@fc_args[0];
  fc_args[0].arg:=DNNL_ARG_SRC;     fc_args[0].memory := pool_dst_memory;
  fc_args[1].arg:=DNNL_ARG_WEIGHTS; fc_args[1].memory := fc_weights_memory;
  fc_args[2].arg:=DNNL_ARG_BIAS;    fc_args[2].memory := fc_bias_memory;
  fc_args[3].arg:=DNNL_ARG_DST;     fc_args[3].memory := fc_dst_memory;
  Inc(net_f);

  // The primitive’s computations are executed on the computational device corresponding to the engine on which the primitive (and memory arguments) were created and happens within the context of the stream.
  CHECK(dnnl_stream_create(@stream, engine, ord(dnnl_stream_default_flags)));
  //////////////////
  // Forward pass
  //////////////////
  // execution of individual primitives
  for i := 0 to net_f-1 do begin
    writeln('Executing (forward) #', i,': ', dnnl_primitive_execute(net_fwd[i].primitive, stream, net_fwd[i].nargs, net_fwd[i].args));
  end;

  writeln('Max Pooling:');
  CHECK(dnnl_memory_get_data_handle(pool_dst_memory, @p));
  for i:=0 to dnnl_memory_desc_get_size(pool_dst_md) div (8*SizeOf(Single)) - 1 do Write(' ', p[i]:6:0);

  writeln;
  writeln('Fully connected: (', dnnl_memory_desc_get_size(fc_dst_md), ' bytes)');
  CHECK(dnnl_memory_get_data_handle(fc_dst_memory, @p));
  for i:=0 to dnnl_memory_desc_get_size(@fc_dst_md) div SizeOf(Single) - 1 do Write(' ', p[i]:6:0);
  writeln;

  {----------------------------------------------------------------------------}
  {-------------------------- Backward Stream ---------------------------------}
  {----------------------------------------------------------------------------}
  net_b := 0;

  pool_diff_src_md := relu_dst_md;
  pool_diff_dst_md := pool_dst_md;

  // create memory for user diff dst data
  CHECK(dnnl_memory_create(@pool_diff_dst_memory, pool_diff_dst_md, engine, DNNL_MEMORY_ALLOCATE));


  // MAX-POOLING
  // The backward propagation computes diff_src based on diff_dst and (in case of max pooling) workspace.
  CHECK(dnnl_pooling_backward_primitive_desc_create(
          @pool_bwd_pd, engine,
          dnnl_pooling_max,
          pool_diff_src_md, pool_diff_dst_md,
          pool_strides, pool_kernel, zero_dilates, zero_padding, nil,
          pool_pd,
          nil));

  // create memory for pool diff src data
  CHECK(dnnl_memory_create(@pool_diff_src_memory, pool_diff_src_md, engine, DNNL_MEMORY_ALLOCATE));

  // finally create backward pooling primitive
  CHECK(dnnl_primitive_create(@pool_bwd, pool_bwd_pd));
  net_bwd[net_b].primitive := pool_bwd;
  net_bwd[net_b].nargs:=3;
  net_bwd[net_b].args:=@pool_bwd_args[0];
  pool_bwd_args[0].arg:=DNNL_ARG_DIFF_DST;  pool_bwd_args[0].memory := pool_diff_dst_memory;
  pool_bwd_args[1].arg:=DNNL_ARG_DIFF_SRC;  pool_bwd_args[1].memory := pool_diff_src_memory;
  pool_bwd_args[2].arg:=DNNL_ARG_WORKSPACE; pool_bwd_args[2].memory := pool_ws_memory;
  Inc(net_b);


  // RELU
  // The backward propagation computes diff_src based on diff_dst and dst
  CHECK(dnnl_eltwise_backward_primitive_desc_create(
         @relu_bwd_pd, engine,
         dnnl_eltwise_relu_use_dst_for_bwd,
         relu_dst_md, relu_dst_md, relu_dst_md,
         0{negative_slope}, 0,
         relu_pd,
         nil));

  // create memory for relu diff src data
  CHECK(dnnl_memory_create(@relu_diff_src_memory, pool_diff_src_md, engine, DNNL_MEMORY_ALLOCATE));

  // finally create backward ReLU primitive
  CHECK(dnnl_primitive_create(@relu_bwd, relu_bwd_pd));
  net_bwd[net_b].primitive := relu_bwd;
  net_bwd[net_b].nargs:=3;
  net_bwd[net_b].args:=@relu_bwd_args[0];
  relu_bwd_args[0].arg:=DNNL_ARG_DST;      relu_bwd_args[0].memory := relu_dst_memory;
  relu_bwd_args[1].arg:=DNNL_ARG_DIFF_DST; relu_bwd_args[1].memory := pool_diff_src_memory;
  relu_bwd_args[2].arg:=DNNL_ARG_DIFF_SRC; relu_bwd_args[2].memory := relu_diff_src_memory;
  Inc(net_b);

  //////////////////
  // Backward pass
  //////////////////
  for i := 0 to net_b-1 do begin
    writeln('Executing (backward) #', i,': ', dnnl_primitive_execute(net_bwd[i].primitive, stream, net_bwd[i].nargs, net_bwd[i].args));
  end;

  CHECK(dnnl_stream_wait(stream));

  CHECK(dnnl_stream_destroy(stream));

  CHECK(dnnl_memory_desc_destroy(input_src_md));
  CHECK(dnnl_memory_desc_destroy(conv_weights_md));
  CHECK(dnnl_memory_desc_destroy(conv_bias_md));
  CHECK(dnnl_memory_desc_destroy(pool_dst_md));
  CHECK(dnnl_memory_desc_destroy(pool_ws_md));

  CHECK(dnnl_memory_destroy(input_src_memory));
  CHECK(dnnl_memory_destroy(conv_dst_memory));
  CHECK(dnnl_memory_destroy(conv_weights_memory));
  CHECK(dnnl_memory_destroy(conv_bias_memory));
  CHECK(dnnl_memory_destroy(relu_dst_memory));
  CHECK(dnnl_memory_destroy(pool_dst_memory));
  CHECK(dnnl_memory_destroy(pool_ws_memory));
  CHECK(dnnl_memory_destroy(fc_weights_memory));
  CHECK(dnnl_memory_destroy(fc_bias_memory));
  CHECK(dnnl_memory_destroy(fc_dst_memory));

  CHECK(dnnl_memory_destroy(pool_diff_dst_memory));
  CHECK(dnnl_memory_destroy(pool_diff_src_memory));
  CHECK(dnnl_memory_destroy(relu_diff_src_memory));

  CHECK(dnnl_primitive_destroy(conv));
  CHECK(dnnl_primitive_destroy(relu));
  CHECK(dnnl_primitive_destroy(pool));
  CHECK(dnnl_primitive_destroy(fc));
  CHECK(dnnl_primitive_destroy(pool_bwd));
  CHECK(dnnl_primitive_destroy(relu_bwd));

  CHECK(dnnl_primitive_desc_destroy(conv_pd));
  CHECK(dnnl_primitive_desc_destroy(relu_pd));
  CHECK(dnnl_primitive_desc_destroy(pool_pd));
  CHECK(dnnl_primitive_desc_destroy(fc_pd));
  CHECK(dnnl_primitive_desc_destroy(pool_bwd_pd));
  CHECK(dnnl_primitive_desc_destroy(relu_bwd_pd));

  CHECK(dnnl_engine_destroy(engine));
{$ENDIF}
  readln;
end.

